from alpaca_trade_api.entity import Asset, BarSet


def list_assets(tradeapi):
    tradeapi.REST().list_assets.return_value = [
        Asset({'asset_class': 'us_equity',
               'exchange': 'NYSE',
               'id': '3ca0202f-01f4-41a0-bb0c-c8864e767ebd',
               'status': 'active',
               'symbol': 'AA',
               'tradable': True}),
        Asset({'asset_class': 'us_equity',
               'exchange': 'NYSE',
               'id': '7595a8d2-68a6-46d7-910c-6b1958491f5c',
               'status': 'active',
               'symbol': 'A',
               'tradable': True}),
    ]


def get_barset(tradeapi):
    tradeapi.REST().get_barset.return_value = BarSet({
        "AA": [{"t": 1544129220,
                "o": 172.26,
                "h": 172.3,
                "l": 172.16,
                "c": 172.18,
                "v": 3892}]})


def list_financials(tradeapi):
    tradeapi.REST().polygon.get.return_value = {
        'AA': [{'symbol': 'AA',
                'reportDate': '2018-06-30T00:00:00.000Z',
                'reportDateStr': '2018-06-30',
                'grossProfit': 947000000,
                'costOfRevenue': 2632000000,
                'operatingRevenue': 3579000000,
                'totalRevenue': 3579000000,
                'operatingIncome': 682000000,
                'netIncome': 75000000,
                'researchAndDevelopment': 9000000,
                'operatingExpense': 265000000,
                'currentAssets': 4281000000,
                'totalAssets': 16518000000,
                'totalLiabilities': None,
                'currentCash': 1095000000,
                'currentDebt': 13000000,
                'totalCash': 1089000000,
                'totalDebt': 1929000000,
                'shareholderEquity': 5024000000,
                'cashChange': -97000000,
                'cashFlow': -430000000,
                'operatingGainsLosses': 38000000},
               {'symbol': 'AA',
                'reportDate': '2018-03-31T00:00:00.000Z',
                'reportDateStr': '2018-03-31',
                'grossProfit': 709000000,
                'costOfRevenue': 2381000000,
                'operatingRevenue': 3090000000,
                'totalRevenue': 3090000000,
                'operatingIncome': 440000000,
                'netIncome': 150000000,
                'researchAndDevelopment': 8000000,
                'operatingExpense': 269000000,
                'currentAssets': 4149000000,
                'totalAssets': 17096000000,
                'totalLiabilities': None,
                'currentCash': 1203000000,
                'currentDebt': 15000000,
                'totalCash': 1196000000,
                'totalDebt': 1460000000,
                'shareholderEquity': 5368000000,
                'cashChange': -166000000,
                'cashFlow': 55000000,
                'operatingGainsLosses': 29000000},
               {'symbol': 'AA',
                'reportDate': '2017-12-31T00:00:00.000Z',
                'reportDateStr': '2017-12-31',
                'grossProfit': 815000000,
                'costOfRevenue': 2359000000,
                'operatingRevenue': 3174000000,
                'totalRevenue': 3174000000,
                'operatingIncome': 531000000,
                'netIncome': -196000000,
                'researchAndDevelopment': 9000000,
                'operatingExpense': 284000000,
                'currentAssets': 4238000000,
                'totalAssets': 17447000000,
                'totalLiabilities': 12924000000,
                'currentCash': 1358000000,
                'currentDebt': 16000000,
                'totalCash': 1358000000,
                'totalDebt': 1404000000,
                'shareholderEquity': 4523000000,
                'cashChange': 232000000,
                'cashFlow': 455000000,
                'operatingGainsLosses': 35000000},
               {'symbol': 'AA',
                'reportDate': '2017-09-30T00:00:00.000Z',
                'reportDateStr': '2017-09-30',
                'grossProfit': 603000000,
                'costOfRevenue': 2361000000,
                'operatingRevenue': 2964000000,
                'totalRevenue': 2964000000,
                'operatingIncome': 331000000,
                'netIncome': 113000000,
                'researchAndDevelopment': 8000000,
                'operatingExpense': 272000000,
                'currentAssets': 3804000000,
                'totalAssets': 17254000000,
                'totalLiabilities': 11392000000,
                'currentCash': 1119000000,
                'currentDebt': 17000000,
                'totalCash': 1119000000,
                'totalDebt': 1401000000,
                'shareholderEquity': 5862000000,
                'cashChange': 175000000,
                'cashFlow': 384000000,
                'operatingGainsLosses': 16000000},
               {'symbol': 'AA',
                'reportDate': '2017-06-30T00:00:00.000Z',
                'reportDateStr': '2017-06-30',
                'grossProfit': 550000000,
                'costOfRevenue': 2309000000,
                'operatingRevenue': 2859000000,
                'totalRevenue': 2859000000,
                'operatingIncome': 280000000,
                'netIncome': 75000000,
                'researchAndDevelopment': 8000000,
                'operatingExpense': 270000000,
                'currentAssets': 3583000000,
                'totalAssets': 16929000000,
                'totalLiabilities': 10975000000,
                'currentCash': 954000000,
                'currentDebt': 19000000,
                'totalCash': 954000000,
                'totalDebt': 1437000000,
                'shareholderEquity': 5954000000,
                'cashChange': 146000000,
                'cashFlow': 311000000,
                'operatingGainsLosses': 46000000}]}


def list_companies(tradeapi):
    tradeapi.REST().polygon.get.return_value = [{
        '_id': '5a77fa81333c4e001174b897',
        'listdate': '1999-11-18',
        'cik': '0001090872',
        'bloomberg': 'EQ0087231700001000',
        'figi': None,
        'lei': 'QUIX8Y7A2WP0XRMW7G29',
        'sic': 3826,
        'country': 'us',
        'industry': 'Medical Diagnostics & Research',
        'sector': 'Healthcare',
        'marketcap': 20626540000,
        'employees': 131,
        'phone': '(408) 345-8886',
        'ceo': 'Michael R. McMullen',
        'url': 'http://www.agilent.com',
        'description': "Agilent Technologies, Inc. provides application focused solutions to the life sciences, diagnostics, and applied chemical markets worldwide. Its Life Sciences and Applied Markets segment offers liquid chromatography systems and components; liquid chromatography mass spectrometry systems; gas chromatography systems and components; gas chromatography mass spectrometry systems; inductively coupled plasma mass spectrometry instruments; atomic absorption instruments; microwave plasma-atomic emission spectrometry instruments; inductively coupled plasma optical emission spectrometry instruments; laboratory software and informatics systems; laboratory automation and robotic systems; dissolution testing; vacuum pumps; and measurement technologies. The company's Diagnostics and Genomics segment provides reagents, instruments, software, and consumables; arrays for DNA mutation detection, genotyping, gene copy number determination, identification of gene rearrangements, DNA methylation profiling, and gene expression profiling, as well as sequencing target enrichment services; and equipment focused on production of synthesized oligonucleotides for use as active pharmaceutical ingredients. Its Agilent CrossLab segment offers GC and LC columns, sample preparation products, custom chemistries, and various laboratory instrument supplies; and startup, operational, training, and compliance support, as well as asset management and consultation services. The company markets and sells its products through direct sales, electronic commerce, resellers, manufacturers' representatives, and distributors. It has a collaboration agreement with University of Leuven to focus on detecting genetic abnormalities in cell-free DNA and embryo biopsies. Agilent Technologies, Inc. was founded in 1999 and is headquartered in Santa Clara, California.",  # noqa
        'exchange': 'New York Stock Exchange',
        'name': 'Agilent Technologies Inc',
        'symbol': 'A',
        '__v': 15,
        'logo': 'https://s3.polygon.io/logos/a/logo.png',
        'created': '2018-02-05T06:32:33.960Z',
        'updated': '2018-02-05T06:32:33.960Z',
        'tags': ['Healthcare',
                 'Diagnostics & Research',
                 'Medical Diagnostics & Research'],
        'similar': ['NATI',
                    'ILMN',
                    'BRKR',
                    'AME',
                    'DHR',
                    'GE',
                    'TER',
                    'TMO',
                    'PKI',
                    'XLV']},
        {'_id': '5a77fa83333c4e001174b898',
         'listdate': '1990-01-02',
         'cik': '0001675149',
         'bloomberg': 'EQ0000000045469815',
         'figi': None,
         'lei': 'ABPN11VOHLHX6QR7XQ48',
         'sic': 3334,
         'country': 'us',
         'industry': 'Metals & Mining',
         'sector': 'Basic Materials',
         'marketcap': 7837461871,
         'employees': 60000,
         'phone': '412-315-2900',
         'ceo': 'Roy C. Harvey',
         'url': 'http://www.alcoa.com',
         'description': 'Alcoa Corporation produces and sells bauxite, alumina, and aluminum products. It operates through six segments, Bauxite, Alumina, Aluminum, Cast Products, Energy, and Rolled Products. The company also offers aluminum cast products; and aluminum sheets for the production of cans for beverage, food, and pet food. In addition, it engages in the generation and sale of renewable energy, as well as provision of ancillary services. The company was formerly known as Alcoa Upstream Corporation and changed its name to Alcoa Corporation in October 2016. Alcoa Corporation is based in New York, New York.',  # noqa
         'exchange': 'New York Stock Exchange',
         'name': 'Alcoa Corp',
         'symbol': 'AA',
         '__v': 19,
         'logo': 'https://s3.polygon.io/logos/aa/logo.png',
         'created': '2018-02-05T06:32:35.464Z',
         'updated': '2018-02-05T06:32:35.464Z',
         'tags': ['Basic Materials', 'Aluminum', 'Metals & Mining'],
         'similar': ['BBL', 'CENX', 'KALU', 'BHP', 'ACH']}]



from uuid import uuid4
from numpy import array
import pandas as pd
from pandas import DataFrame
from six import (
    iteritems,
)
from toolz import groupby, juxt
from toolz.curried.operator import getitem

from zipline.lib.adjusted_array import ensure_adjusted_array, ensure_ndarray
from zipline.errors import NoFurtherDataError
from zipline.pipeline.engine import default_populate_initial_workspace
from zipline.pipeline.term import AssetExists, InputDates, LoadableTerm
from zipline.utils.calendars import get_calendar
from zipline.utils.numpy_utils import (
    as_column,
    repeat_first_axis,
)
from zipline.utils.pandas_utils import explode


class LivePipelineEngine(object):

    def __init__(self,
                 list_symbols,
                 calendar=None,
                 populate_initial_workspace=None):
        self._list_symbols = list_symbols
        if calendar is None:
            calendar = get_calendar('NYSE').all_sessions
        self._calendar = calendar

        self._root_mask_term = AssetExists()
        self._root_mask_dates_term = InputDates()

        self._populate_initial_workspace = (
            populate_initial_workspace or default_populate_initial_workspace
        )

    def run_pipeline(self, pipeline):
        now = pd.Timestamp.now(tz=self._calendar.tz)
        today = pd.Timestamp(
            year=now.year, month=now.month, day=now.day,
            tz='utc')

        end_date = self._calendar[self._calendar.get_loc(
            today, method='ffill'
        )]
        start_date = end_date
        screen_name = uuid4().hex
        graph = pipeline.to_execution_plan(screen_name,
                                           self._root_mask_term,
                                           self._calendar,
                                           start_date,
                                           end_date,
                                           )
        extra_rows = graph.extra_rows[self._root_mask_term]
        root_mask = self._compute_root_mask(start_date, end_date, extra_rows)
        dates, assets, root_mask_values = explode(root_mask)

        initial_workspace = self._populate_initial_workspace(
            {
                self._root_mask_term: root_mask_values,
                self._root_mask_dates_term: as_column(dates.values)
            },
            self._root_mask_term,
            graph,
            dates,
            assets,
        )

        results = self.compute_chunk(
            graph,
            dates,
            assets,
            initial_workspace,
        )

        return self._to_narrow(
            graph.outputs,
            results,
            results.pop(screen_name),
            dates[extra_rows:],
            assets,
        )

    def _compute_root_mask(self, start_date, end_date, extra_rows):
        """
        Compute a lifetimes matrix from our AssetFinder, then drop columns that
        didn't exist at all during the query dates.

        Parameters
        ----------
        start_date : pd.Timestamp
            Base start date for the matrix.
        end_date : pd.Timestamp
            End date for the matrix.
        extra_rows : int
            Number of extra rows to compute before `start_date`.
            Extra rows are needed by terms like moving averages that require a
            trailing window of data.

        Returns
        -------
        lifetimes : pd.DataFrame
            Frame of dtype `bool` containing dates from `extra_rows` days
            before `start_date`, continuing through to `end_date`.  The
            returned frame contains as columns all assets in our AssetFinder
            that existed for at least one day between `start_date` and
            `end_date`.
        """
        calendar = self._calendar
        start_idx, end_idx = calendar.slice_locs(start_date, end_date)
        if start_idx < extra_rows:
            raise NoFurtherDataError.from_lookback_window(
                initial_message="Insufficient data to compute Pipeline:",
                first_date=calendar[0],
                lookback_start=start_date,
                lookback_length=extra_rows,
            )

        # Build lifetimes matrix reaching back to `extra_rows` days before
        # `start_date.`
        symbols = self._list_symbols()
        dates = calendar[start_idx - extra_rows:end_idx]
        symbols = sorted(symbols)
        lifetimes = pd.DataFrame(True, index=dates, columns=symbols)

        assert lifetimes.index[extra_rows] == start_date
        assert lifetimes.index[-1] == end_date
        if not lifetimes.columns.unique:
            columns = lifetimes.columns
            duplicated = columns[columns.duplicated()].unique()
            raise AssertionError("Duplicated sids: %d" % duplicated)

        # Filter out columns that didn't exist between the requested start and
        # end dates.
        existed = lifetimes.iloc[extra_rows:].any()
        ret = lifetimes.loc[:, existed]
        shape = ret.shape
        assert shape[0] * shape[1] != 0, 'root mask cannot be empty'
        return ret

    @staticmethod
    def _inputs_for_term(term, workspace, graph):
        """
        Compute inputs for the given term.

        This is mostly complicated by the fact that for each input we store as
        many rows as will be necessary to serve **any** computation requiring
        that input.
        """
        offsets = graph.offset
        out = []
        if term.windowed:
            # If term is windowed, then all input data should be instances of
            # AdjustedArray.
            for input_ in term.inputs:
                adjusted_array = ensure_adjusted_array(
                    workspace[input_], input_.missing_value,
                )
                out.append(
                    adjusted_array.traverse(
                        window_length=term.window_length,
                        offset=offsets[term, input_],
                    )
                )
        else:
            # If term is not windowed, input_data may be an AdjustedArray or
            # np.ndarray.  Coerce the former to the latter.
            for input_ in term.inputs:
                input_data = ensure_ndarray(workspace[input_])
                offset = offsets[term, input_]
                # OPTIMIZATION: Don't make a copy by doing input_data[0:] if
                # offset is zero.
                if offset:
                    input_data = input_data[offset:]
                out.append(input_data)
        return out

    def compute_chunk(self, graph, dates, symbols, initial_workspace):
        """
        Compute the Pipeline terms in the graph for the requested start and end
        dates.

        Parameters
        ----------
        graph : zipline.pipeline.graph.TermGraph
        dates : pd.DatetimeIndex
            Row labels for our root mask.
        symbols : list
            Column labels for our root mask.
        initial_workspace : dict
            Map from term -> output.
            Must contain at least entry for `self._root_mask_term` whose shape
            is `(len(dates), len(symbols))`, but may contain additional
            pre-computed terms for testing or optimization purposes.

        Returns
        -------
        results : dict
            Dictionary mapping requested results to outputs.
        """
        self._validate_compute_chunk_params(dates, symbols, initial_workspace)
        # get_loader = self.get_loader

        # Copy the supplied initial workspace so we don't mutate it in place.
        workspace = initial_workspace.copy()

        # If loadable terms share the same loader and extra_rows, load them all
        # together.
        loader_group_key = juxt(
            lambda x: x.dataset.get_loader(), getitem(
                graph.extra_rows))
        loader_groups = groupby(loader_group_key, graph.loadable_terms)

        refcounts = graph.initial_refcounts(workspace)

        for term in graph.execution_order(refcounts):
            # `term` may have been supplied in `initial_workspace`, and in the
            # future we may pre-compute loadable terms coming from the same
            # dataset.  In either case, we will already have an entry for this
            # term, which we shouldn't re-compute.
            if term in workspace:
                continue

            # Asset labels are always the same, but date labels vary by how
            # many extra rows are needed.
            mask, mask_dates = graph.mask_and_dates_for_term(
                term,
                self._root_mask_term,
                workspace,
                dates,
            )

            if isinstance(term, LoadableTerm):
                to_load = sorted(
                    loader_groups[loader_group_key(term)],
                    key=lambda t: t.dataset
                )
                loader = term.dataset.get_loader()
                loaded = loader.load_adjusted_array(
                    to_load, mask_dates, symbols, mask,
                )
                workspace.update(loaded)
            else:
                workspace[term] = term._compute(
                    self._inputs_for_term(term, workspace, graph),
                    mask_dates,
                    symbols,
                    mask,
                )
                if term.ndim == 2:
                    assert workspace[term].shape == mask.shape
                else:
                    assert workspace[term].shape == (mask.shape[0], 1)

                # Decref dependencies of ``term``, and clear any terms whose
                # refcounts hit 0.
                for garbage_term in graph.decref_dependencies(term, refcounts):
                    del workspace[garbage_term]

        out = {}
        graph_extra_rows = graph.extra_rows
        for name, term in iteritems(graph.outputs):
            # Truncate off extra rows from outputs.
            out[name] = workspace[term][graph_extra_rows[term]:]
        return out

    def _to_narrow(self, terms, data, mask, dates, symbols):
        """
        Convert raw computed pipeline results into a DataFrame for public APIs.

        Parameters
        ----------
        terms : dict[str -> Term]
            Dict mapping column names to terms.
        data : dict[str -> ndarray[ndim=2]]
            Dict mapping column names to computed results for those names.
        mask : ndarray[bool, ndim=2]
            Mask array of values to keep.
        dates : ndarray[datetime64, ndim=1]
            Row index for arrays `data` and `mask`
        symbols : list
            Column index

        Returns
        -------
        results : pd.DataFrame
            The indices of `results` are as follows:

            index : two-tiered MultiIndex of (date, asset).
                Contains an entry for each (date, asset) pair corresponding to
                a `True` value in `mask`.
            columns : Index of str
                One column per entry in `data`.

        If mask[date, asset] is True, then result.loc[(date, asset), colname]
        will contain the value of data[colname][date, asset].
        """
        assert len(dates) == 1
        if not mask.any():
            # Manually handle the empty DataFrame case. This is a workaround
            # to pandas failing to tz_localize an empty dataframe with a
            # MultiIndex. It also saves us the work of applying a known-empty
            # mask to each array.
            #
            # Slicing `dates` here to preserve pandas metadata.
            # empty_dates = dates[:0]
            empty_assets = array([], dtype=object)
            return DataFrame(
                data={
                    name: array([], dtype=arr.dtype)
                    for name, arr in iteritems(data)
                },
                index=pd.Index(empty_assets),
                # index=MultiIndex.from_arrays([empty_dates, empty_assets]),
            )

        # resolved_assets = array(self._finder.retrieve_all(assets))
        # dates_kept = repeat_last_axis(dates.values, len(symbols))[mask]
        # assets_kept = repeat_first_axis(resolved_assets, len(dates))[mask]
        assets_kept = repeat_first_axis(symbols, len(dates))[mask]

        final_columns = {}
        for name in data:
            # Each term that computed an output has its postprocess method
            # called on the filtered result.
            #
            # As of Mon May 2 15:38:47 2016, we only use this to convert
            # LabelArrays into categoricals.
            final_columns[name] = terms[name].postprocess(data[name][mask])

        return DataFrame(
            data=final_columns,
            index=pd.Index(assets_kept),
            # index=MultiIndex.from_arrays([dates_kept, assets_kept]),
        )

    def _validate_compute_chunk_params(
            self, dates, symbols, initial_workspace):
        """
        Verify that the values passed to compute_chunk are well-formed.
        """
        root = self._root_mask_term
        clsname = type(self).__name__

        # Writing this out explicitly so this errors in testing if we change
        # the name without updating this line.
        compute_chunk_name = self.compute_chunk.__name__
        if root not in initial_workspace:
            raise AssertionError(
                "root_mask values not supplied to {cls}.{method}".format(
                    cls=clsname,
                    method=compute_chunk_name,
                )
            )

        shape = initial_workspace[root].shape
        implied_shape = len(dates), len(symbols)
        if shape != implied_shape:
            raise AssertionError(
                "root_mask shape is {shape}, but received dates/symbols "
                "imply that shape should be {implied}".format(
                    shape=shape,
                    implied=implied_shape,
                )
            )



import alpaca_trade_api as tradeapi

from .util import (
    daily_cache, parallelize
)


def list_symbols():
    return [
        a.symbol for a in tradeapi.REST().list_assets()
        if a.tradable and a.status == 'active'
    ]


def company():
    all_symbols = list_symbols()
    return _company(all_symbols)


@daily_cache(filename='polygon_company.pkl')
def _company(all_symbols):
    def fetch(symbols):
        api = tradeapi.REST()
        params = {
            'symbols': ','.join(symbols),
        }
        response = api.polygon.get('/meta/symbols/company', params=params)
        return {
            o['symbol']: o for o in response
        }

    return parallelize(fetch, workers=25, splitlen=50)(all_symbols)


def financials():
    all_symbols = list_symbols()
    return _financials(all_symbols)


@daily_cache(filename='polygon_financials.pkl')
def _financials(all_symbols):
    def fetch(symbols):
        api = tradeapi.REST()
        params = {
            'symbols': ','.join(symbols),
        }
        return api.polygon.get('/meta/symbols/financials', params=params)

    return parallelize(fetch, workers=25, splitlen=50)(all_symbols)



import numpy as np

from zipline.pipeline.loaders.base import PipelineLoader

from pipeline_live.data.sources import polygon


class PolygonCompanyLoader(PipelineLoader):

    def load_adjusted_array(self, columns, dates, symbols, mask):

        company = polygon.company()
        out = {}
        for c in columns:
            data = [
                company.get(symbol, {}).get(c.name, c.missing_value)
                for symbol in symbols
            ]
            out[c] = np.tile(np.array(data, dtype=c.dtype), (len(dates), 1))

        return out



'''
Duplicate builtin factor classes in zipline with IEX's USEquityPricing
'''

from zipline.pipeline.data import USEquityPricing as z_pricing
from zipline.pipeline import factors as z_factors

from .pricing import USEquityPricing as alpaca_pricing


def _replace_inputs(inputs):
    map = {
        z_pricing.open: alpaca_pricing.open,
        z_pricing.high: alpaca_pricing.high,
        z_pricing.low: alpaca_pricing.low,
        z_pricing.close: alpaca_pricing.close,
        z_pricing.volume: alpaca_pricing.volume,
    }

    if type(inputs) not in (list, tuple, set):
        return inputs
    return tuple([
        map.get(inp, inp) for inp in inputs
    ])


for name in dir(z_factors):
    factor = getattr(z_factors, name)
    if factor != z_factors.Factor and hasattr(
            factor, 'inputs') and issubclass(
            factor, z_factors.Factor):
        new_factor = type(factor.__name__, (factor,), {
            'inputs': _replace_inputs(factor.inputs)
        })
        locals()[factor.__name__] = new_factor



import pytest
import tempfile
from unittest.mock import patch

from pipeline_live.data.sources import polygon as sources_polygon
from pipeline_live.data.sources import iex as sources_iex
from pipeline_live.data.sources import alpaca as sources_alpaca


@pytest.fixture
def tradeapi():
    with patch.object(sources_polygon, 'tradeapi') as tradeapi:
        yield tradeapi


@pytest.fixture
def alpaca_tradeapi():
    with patch.object(sources_alpaca, 'tradeapi') as tradeapi:
        yield tradeapi


@pytest.fixture
def refdata():
    with patch.object(sources_iex, 'refdata') as refdata:
        yield refdata


@pytest.fixture
def stocks():
    with patch.object(sources_iex, 'Stock') as stocks:
        yield stocks


@pytest.fixture
def data_path():
    with patch('zipline.utils.paths.data_path') as data_path:
        with tempfile.TemporaryDirectory() as t:
            data_path.return_value = t
            yield data_path



from iexfinance import refdata
from iexfinance.stocks import Stock
import pandas as pd

from .util import (
    daily_cache, parallelize
)


def list_symbols():
    return [
        symbol['symbol'] for symbol in refdata.get_iex_symbols()
    ]


def _ensure_dict(result, symbols):
    if len(symbols) == 1:
        return {symbols[0]: result}
    return result


class IEXGetter(object):

    def __init__(self, method):
        self._method = method

    def __call__(self):
        method_name = 'get_{}'.format(self._method)

        @daily_cache(filename='iex_{}.pkl'.format(self._method))
        def _get(all_symbols):
            def fetch(symbols):
                return _ensure_dict(
                    getattr(Stock(symbols), method_name)(),
                    symbols
                )

            return parallelize(fetch, splitlen=99)(all_symbols)

        all_symbols = list_symbols()
        return _get(all_symbols)


key_stats = IEXGetter('key_stats')
company = IEXGetter('company')
financials = IEXGetter('financials')


def get_stockprices(chart_range='1y'):
    '''
    This is a proxy to the main fetch function to cache
    the result based on the chart range parameter.
    '''

    all_symbols = list_symbols()

    @daily_cache(filename='iex_chart_{}'.format(chart_range))
    def get_stockprices_cached(all_symbols):
        return _get_stockprices(all_symbols, chart_range)

    return get_stockprices_cached(all_symbols)


def _get_stockprices(symbols, chart_range='1y'):
    '''Get stock data (key stats and previous) from IEX.
    Just deal with IEX's 100 stocks limit per request.
    '''

    def fetch(symbols):
        charts = _ensure_dict(
            Stock(symbols).get_chart(range=chart_range),
            symbols
        )
        result = {}
        for symbol, obj in charts.items():
            df = pd.DataFrame(
                obj,
                columns=('date', 'open', 'high', 'low', 'close', 'volume'),
            ).set_index('date')
            df.index = pd.to_datetime(df.index, utc=True)
            result[symbol] = df
        return result

    return parallelize(fetch, splitlen=99)(symbols)



import pandas as pd
import numpy as np

from .datamock import mock_tradeapi

from pipeline_live.data.alpaca.pricing import USEquityPricing

def test_pricing_loader(refdata, alpaca_tradeapi, data_path):
    mock_tradeapi.list_assets(alpaca_tradeapi)
    mock_tradeapi.get_barset(alpaca_tradeapi)

    loader = USEquityPricing.get_loader()
    columns = [USEquityPricing.close]
    dates = [pd.Timestamp('2018-08-22', tz='UTC')]
    symbols = ['AA']
    mask = np.zeros((1, 1), dtype='bool')
    out = loader.load_adjusted_array(columns, dates, symbols, mask)

    assert out[USEquityPricing.close]._data.shape == (1, 1)


import pandas as pd
from pipeline_live.data.polygon.fundamentals import PolygonCompany
from pipeline_live.data.polygon.filters import (
    IsPrimaryShareEmulation, StaticSymbols
)

from .datamock import mock_tradeapi


def test_PolygonCompany(tradeapi, data_path):
    mock_tradeapi.list_assets(tradeapi)
    mock_tradeapi.list_companies(tradeapi)

    marketcap = PolygonCompany.marketcap
    loader = marketcap.dataset.get_loader()
    date = pd.Timestamp('2018-08-20', tz='utc')
    out = loader.load_adjusted_array([marketcap], [date], ['AA'], [])

    assert out[marketcap][0][0] > 10e6


def test_IsPrimaryShareEmulation(tradeapi, data_path):
    mock_tradeapi.list_companies(tradeapi)
    mock_tradeapi.list_financials(tradeapi)

    emu = IsPrimaryShareEmulation()

    today = object()
    symbols = ['A', 'AA']
    out = [None, None]
    emu.compute(today, symbols, out)
    assert not out[0]


def test_StaticSymbols(tradeapi, data_path):
    symbols = ('A', 'BB')
    emu = StaticSymbols(symbols=symbols)

    today = object()
    assets = ['A', 'AA', 'BB']
    out = [None, None, None]
    emu.compute(today, assets, out, symbols)
    assert not out[1]



import numpy as np

from pipeline_live.data.sources import polygon
from zipline.pipeline.filters import CustomFilter


class IsPrimaryShareEmulation(CustomFilter):
    inputs = ()
    window_length = 1

    def compute(self, today, symbols, out, *inputs):
        company = polygon.company()
        financials = polygon.financials()
        ary = np.array([
            symbol in company and
            'country' in company[symbol] and
            [company[symbol].get('country') == 'us'] and
            symbol in financials and
            len(financials[symbol]) > 0 and
            financials[symbol][0].get('totalRevenue') is not None
            for symbol in symbols
        ], dtype=bool)
        out[:] = ary


class StaticSymbols(CustomFilter):
    inputs = ()
    window_length = 1
    params = ('symbols',)

    def compute(self, today, assets, out, symbols, *inputs):
        ary = np.array([symbol in symbols for symbol in assets])
        out[:] = ary



import alpaca_trade_api as tradeapi

from .util import (
    daily_cache, parallelize
)


def list_symbols():
    return [
        a.symbol for a in tradeapi.REST().list_assets()
        if a.tradable and a.status == 'active'
    ]


def get_stockprices(limit=365, timespan='day'):
    all_symbols = list_symbols()

    @daily_cache(filename='alpaca_chart_{}'.format(limit))
    def get_stockprices_cached(all_symbols):
        return _get_stockprices(all_symbols, limit, timespan)

    return get_stockprices_cached(all_symbols)


def _get_stockprices(symbols, limit=365, timespan='day'):
    '''Get stock data (key stats and previous) from Alpaca.
    Just deal with Alpaca's 200 stocks per request limit.
    '''

    def fetch(symbols):
        barset = tradeapi.REST().get_barset(symbols, timespan, limit)
        data = {}
        for symbol in barset:
            df = barset[symbol].df
            # Update the index format for comparison with the trading calendar
            df.index = df.index.tz_convert('UTC').normalize()
            data[symbol] = df.asfreq('C')

        return data

    return parallelize(fetch, splitlen=199)(symbols)



import concurrent.futures
import hashlib
import logbook
import pandas as pd
import os
import pickle

from zipline.utils import paths


log = logbook.Logger(__name__)


def parallelize(mapfunc, workers=10, splitlen=10):

    def wrapper(symbols):
        result = {}
        with concurrent.futures.ThreadPoolExecutor(
                max_workers=workers) as executor:
            tasks = []
            for i in range(0, len(symbols), splitlen):
                part = symbols[i:i + splitlen]
                task = executor.submit(mapfunc, part)
                tasks.append(task)

            total_count = len(symbols)
            report_percent = 10
            processed = 0
            for task in concurrent.futures.as_completed(tasks):
                task_result = task.result()
                result.update(task_result)
                processed += len(task_result)
                percent = processed / total_count * 100
                if percent >= report_percent:
                    log.debug('{}: {:.2f}% completed'.format(
                        mapfunc.__name__, percent))
                    report_percent = (percent + 10.0) // 10 * 10
        return result

    return wrapper


def daily_cache(filename):
    kwd_mark = '||'

    def decorator(func):
        def wrapper(*args, **kwargs):
            key = args + (kwd_mark,) + tuple(sorted(kwargs.items()))
            hash = hashlib.md5()
            hash.update(str(key).encode('utf-8'))
            hash.update(pd.Timestamp.utcnow().strftime(
                '%Y-%m-%d').encode('utf-8'))
            digest = hash.hexdigest()
            dirpath = paths.data_path(['dailycache'])
            os.makedirs(dirpath, exist_ok=True)
            filepath = os.path.join(dirpath, filename)
            if os.path.exists(filepath):
                try:
                    with open(filepath, 'rb') as fp:
                        ret = pickle.load(fp)
                        if ret['digest'] == digest:
                            return ret['body']
                        print('{}: digest mismatch {} != {}, reloading'.format(
                            filepath, ret['digest'], digest
                        ))
                except Exception as e:
                    print('cache error {}'.format(e))
            body = func(*args, **kwargs)

            with open(filepath, 'wb') as fp:
                pickle.dump({
                    'digest': digest,
                    'body': body,
                }, fp)
            return body
        return wrapper
    return decorator



import numpy as np

from pipeline_live.data.sources import iex
from zipline.pipeline.loaders.base import PipelineLoader
from zipline.utils.numpy_utils import object_dtype


class IEXBaseLoader(PipelineLoader):

    def load_adjusted_array(self, columns, dates, symbols, mask):
        symbol_dict = self._load()
        out = {}
        for c in columns:
            data = np.array([
                symbol_dict.get(symbol, {}).get(c.name, c.missing_value)
                for symbol in symbols
            ], dtype=c.dtype)
            if c.dtype == object_dtype:
                data[data == None] = c.missing_value  # noqa
            out[c] = np.tile(data, (len(dates), 1))
        return out


class IEXKeyStatsLoader(IEXBaseLoader):

    def _load(self):
        return iex.key_stats()


class IEXCompanyLoader(IEXBaseLoader):

    def _load(self):
        return iex.company()





def get_available_symbols(refdata):
    refdata.get_iex_symbols.return_value = [
        {
            "symbol": "A",
            "name": "AGILENT TECHNOLOGIES INC",
            "date": "2017-04-19",
            "isEnabled": True,
            "type": "cs",
            "iexId": "1"
        },
        {
            "symbol": "AA",
            "name": "ALCOA CORP",
            "date": "2017-04-19",
            "isEnabled": True,
            "type": "cs",
            "iexId": "12042"
        }
    ]


def get_key_stats(stocks):
    stocks().get_key_stats.return_value = {
        'A': {
            'companyName': 'Agilent Technologies Inc.',
            'marketcap': 20626540000,
            'beta': 1.369553,
            'week52high': 75,
            'week52low': 60.42,
            'week52change': 7.1948,
            'shortInterest': 5255310,
            'shortDate': '2018-07-31',
            'dividendRate': 0.596,
            'dividendYield': 0.9217445,
            'exDividendDate': '2018-07-02 00:00:00.0',
            'latestEPS': 2.1,
            'latestEPSDate': '2017-10-31',
            'sharesOutstanding': 319000000,
            'float': 315254758,
            'returnOnEquity': 6.5,
            'consensusEPS': 0.65,
            'numberOfEstimates': 10,
            'EPSSurpriseDollar': None,
            'EPSSurprisePercent': 0,
            'symbol': 'A',
            'EBITDA': 865000000,
            'revenue': 3514000000,
            'grossProfit': 1916000000,
            'cash': 8128000000,
            'debt': 6237000000,
            'ttmEPS': 2.57,
            'revenuePerShare': 11,
            'revenuePerEmployee': 251000,
            'peRatioHigh': 48.9,
            'peRatioLow': 7.8,
            'returnOnAssets': 3.59,
            'returnOnCapital': None,
            'profitMargin': 6.2,
            'priceToSales': 4.344874,
            'priceToBook': 4.47,
            'day200MovingAvg': 66.93319,
            'day50MovingAvg': 64.37124,
            'institutionPercent': 91.3,
            'insiderPercent': 1.2,
            'shortRatio': 2.8348546,
            'year5ChangePercent': 1.0195710363653967,
            'year2ChangePercent': 0.3607160068267285,
            'year1ChangePercent': 0.07194782502018399,
            'ytdChangePercent': -0.039041716824944034,
            'month6ChangePercent': -0.08608936423423595,
            'month3ChangePercent': 0.019282183212267934,
            'month1ChangePercent': 0.025697969543147167,
            'day5ChangePercent': -0.0194115862905672,
            'day30ChangePercent': 0.024397972116603283},
        'AA': {
            'companyName': 'Alcoa Corporation',
            'marketcap': 7837461871,
            'beta': 0.596731,
            'week52high': 62.35,
            'week52low': 38.05,
            'week52change': 11.1905,
            'shortInterest': 9035385,
            'shortDate': '2018-07-31',
            'dividendRate': 0,
            'dividendYield': 0,
            'exDividendDate': 0,
            'latestEPS': 1.16,
            'latestEPSDate': '2017-12-30',
            'sharesOutstanding': 186473040,
            'float': 186017718,
            'returnOnEquity': 2.59,
            'consensusEPS': 1.33,
            'numberOfEstimates': 5,
            'EPSSurpriseDollar': None,
            'EPSSurprisePercent': 14.2857,
            'symbol': 'AA',
            'EBITDA': 940000000,
            'revenue': 6138000000,
            'grossProfit': 1418000000,
            'cash': 1119000000,
            'debt': 1401000000,
            'ttmEPS': 4.05,
            'revenuePerShare': 33,
            'revenuePerEmployee': 420411,
            'peRatioHigh': 0,
            'peRatioLow': 0,
            'returnOnAssets': 0.85,
            'returnOnCapital': None,
            'profitMargin': 4.82,
            'priceToSales': 0.6035221,
            'priceToBook': 1.56,
            'day200MovingAvg': 47.8935,
            'day50MovingAvg': 45.0584,
            'institutionPercent': 86.9,
            'insiderPercent': 0.2,
            'shortRatio': 2.0121994,
            'year5ChangePercent': 0.8747240101162838,
            'year2ChangePercent': 0.8747240101162838,
            'year1ChangePercent': 0.11190476190476202,
            'ytdChangePercent': -0.23817292006525287,
            'month6ChangePercent': -0.1256500936134803,
            'month3ChangePercent': -0.17116939459672648,
            'month1ChangePercent': -0.12473969179508541,
            'day5ChangePercent': -0.05444319460067495,
            'day30ChangePercent': -0.12963346448540067}}


def get_financials(stocks):
    stocks().get_financials.return_value = {'AA': [
        {'reportDate': '2018-06-30',
         'grossProfit': 947000000,
         'costOfRevenue': 2632000000,
         'operatingRevenue': 3579000000,
         'totalRevenue': 3579000000,
         'operatingIncome': 682000000,
         'netIncome': 75000000,
         'researchAndDevelopment': 9000000,
         'operatingExpense': 265000000,
         'currentAssets': 4281000000,
         'totalAssets': 16518000000,
         'totalLiabilities': None,
         'currentCash': 1095000000,
         'currentDebt': 13000000,
         'totalCash': 1089000000,
         'totalDebt': 1929000000,
         'shareholderEquity': 5024000000,
         'cashChange': -97000000,
         'cashFlow': -430000000,
         'operatingGainsLosses': 38000000},
        {'reportDate': '2018-03-31',
            'grossProfit': 709000000,
            'costOfRevenue': 2381000000,
            'operatingRevenue': 3090000000,
            'totalRevenue': 3090000000,
            'operatingIncome': 440000000,
            'netIncome': 150000000,
            'researchAndDevelopment': 8000000,
            'operatingExpense': 269000000,
            'currentAssets': 4149000000,
            'totalAssets': 17096000000,
            'totalLiabilities': None,
            'currentCash': 1203000000,
            'currentDebt': 15000000,
            'totalCash': 1196000000,
            'totalDebt': 1460000000,
            'shareholderEquity': 5368000000,
            'cashChange': -166000000,
            'cashFlow': 55000000,
            'operatingGainsLosses': 29000000},
        {'reportDate': '2017-12-31',
            'grossProfit': 815000000,
            'costOfRevenue': 2359000000,
            'operatingRevenue': 3174000000,
            'totalRevenue': 3174000000,
            'operatingIncome': 531000000,
            'netIncome': -196000000,
            'researchAndDevelopment': 9000000,
            'operatingExpense': 284000000,
            'currentAssets': 4238000000,
            'totalAssets': 17447000000,
            'totalLiabilities': None,
            'currentCash': 1358000000,
            'currentDebt': 16000000,
            'totalCash': 1358000000,
            'totalDebt': 1404000000,
            'shareholderEquity': 4523000000,
            'cashChange': 232000000,
            'cashFlow': 455000000,
            'operatingGainsLosses': 35000000},
        {'reportDate': '2017-09-30',
            'grossProfit': 603000000,
            'costOfRevenue': 2361000000,
            'operatingRevenue': 2964000000,
            'totalRevenue': 2964000000,
            'operatingIncome': 331000000,
            'netIncome': 113000000,
            'researchAndDevelopment': 8000000,
            'operatingExpense': 272000000,
            'currentAssets': 3804000000,
            'totalAssets': 17254000000,
            'totalLiabilities': None,
            'currentCash': 1119000000,
            'currentDebt': 17000000,
            'totalCash': 1119000000,
            'totalDebt': 1401000000,
            'shareholderEquity': 5862000000,
            'cashChange': 165000000,
            'cashFlow': 384000000,
            'operatingGainsLosses': 16000000}]}


def get_chart(stocks):
    stocks().get_chart.return_value = {
        'AA': [{
            'date': '2018-08-21',
            'open': 41.88,
            'high': 43.22,
            'low': 41.88,
            'close': 43.03,
            'volume': 3095888,
            'unadjustedVolume': 3095888,
            'change': 1.2,
            'changePercent': 2.869,
            'vwap': 42.8847,
            'label': 'Aug 21',
            'changeOverTime': 0.04113234938301483},
            {'date': '2018-08-22',
             'open': 43.2,
             'high': 43.79,
             'low': 43.12,
             'close': 43.22,
             'volume': 2037865,
             'unadjustedVolume': 2037865,
             'change': 0.19,
             'changePercent': 0.442,
             'vwap': 43.3733,
             'label': 'Aug 22',
             'changeOverTime': 0.0457294943140576}]}



import numpy as np

from zipline.utils.numpy_utils import (
    object_dtype, datetime64D_dtype, float64_dtype,
)
from zipline.pipeline.data.dataset import Column, DataSet

from .fundamentals_loader import IEXKeyStatsLoader, IEXCompanyLoader


class IEXKeyStats(DataSet):

    '''
  "companyName": "Apple Inc.",
  "marketcap": 760334287200,
  "beta": 1.295227,
  "week52high": 156.65,
  "week52low": 93.63,
  "week52change": 58.801903,
  "shortInterest": 55544287,
  "shortDate": "2017-06-15",
  "dividendRate": 2.52,
  "dividendYield": 1.7280395,
  "exDividendDate": "2017-05-11 00:00:00.0",
  "latestEPS": 8.29,
  "latestEPSDate": "2016-09-30",
  "sharesOutstanding": 5213840000,
  "float": 5203997571, "returnOnEquity": 0.08772939519857577,
  "consensusEPS": 3.22,
  "numberOfEstimates": 15,
  "symbol": "AAPL",
  "EBITDA": 73828000000,
  "revenue": 220457000000,
  "grossProfit": 84686000000,
  "cash": 256464000000,
  "debt": 358038000000,
  "ttmEPS": 8.55,
  "revenuePerShare": 42.2830389885382,
  "revenuePerEmployee": 1900491.3793103448, "peRatioHigh": 25.5,
  "peRatioLow": 8.7,
  "EPSSurpriseDollar": null,
  "EPSSurprisePercent": 3.9604,
  "returnOnAssets": 14.15,
  "returnOnCapital": null,
  "profitMargin": 20.73,
  "priceToSales": 3.6668503,
  "priceToBook": 6.19,
  "day200MovingAvg": 140.60541,
  "day50MovingAvg": 156.49678,
  "institutionPercent": 32.1,
  "insiderPercent": null,
  "shortRatio": 1.6915414,
  "year5ChangePercent": 0.5902546932200027,
  "year2ChangePercent": 0.3777449874142869,
  "year1ChangePercent": 0.39751716851558366,
  "ytdChangePercent": 0.36659492036160124,
  "month6ChangePercent": 0.12208398133748043,
  "month3ChangePercent": 0.08466584665846649,
  "month1ChangePercent": 0.009668596145283263,
  "day5ChangePercent": -0.005762605699968781
    '''

    companyName = Column(object_dtype)
    marketcap = Column(float64_dtype, missing_value=np.nan)
    beta = Column(float64_dtype, missing_value=np.nan)
    week52high = Column(float64_dtype, missing_value=np.nan)
    week52low = Column(float64_dtype, missing_value=np.nan)
    week52change = Column(float64_dtype, missing_value=np.nan)
    shortInterest = Column(float64_dtype, missing_value=np.nan)
    shortDate = Column(
        datetime64D_dtype,
        missing_value=np.datetime64('1970-01-01'))
    dividendRate = Column(float64_dtype, missing_value=np.nan)
    dividendYield = Column(float64_dtype, missing_value=np.nan)
    exDividendDate = Column(
        datetime64D_dtype,
        missing_value=np.datetime64('1970-01-01'))
    latestEPS = Column(float64_dtype, missing_value=np.nan)
    latestEPSDate = Column(
        datetime64D_dtype,
        missing_value=np.datetime64('1970-01-01'))
    sharesOutstanding = Column(float64_dtype, missing_value=np.nan)
    float = Column(float64_dtype, missing_value=np.nan)
    returnOnEquity = Column(float64_dtype, missing_value=np.nan)
    consensusEPS = Column(float64_dtype, missing_value=np.nan)
    numberOfEstimates = Column(float64_dtype, missing_value=np.nan)
    symbol = Column(object_dtype)
    EBITDA = Column(float64_dtype, missing_value=np.nan)
    revenue = Column(float64_dtype, missing_value=np.nan)
    grossProfit = Column(float64_dtype, missing_value=np.nan)
    cash = Column(float64_dtype, missing_value=np.nan)
    debt = Column(float64_dtype, missing_value=np.nan)
    ttmEPS = Column(float64_dtype, missing_value=np.nan)
    revenuePerShare = Column(float64_dtype, missing_value=np.nan)
    revenuePerEmployee = Column(float64_dtype, missing_value=np.nan)
    peRatioHigh = Column(float64_dtype, missing_value=np.nan)
    peRatioLow = Column(float64_dtype, missing_value=np.nan)
    EPSSurpriseDollar = Column(float64_dtype, missing_value=np.nan)
    EPSSurprisePercent = Column(float64_dtype, missing_value=np.nan)
    returnOnAssets = Column(float64_dtype, missing_value=np.nan)
    returnOnCapital = Column(float64_dtype, missing_value=np.nan)
    profitMargin = Column(float64_dtype, missing_value=np.nan)
    priceToSales = Column(float64_dtype, missing_value=np.nan)
    priceToBook = Column(float64_dtype, missing_value=np.nan)
    day200MovingAvg = Column(float64_dtype, missing_value=np.nan)
    day50MovingAvg = Column(float64_dtype, missing_value=np.nan)
    institutionPercent = Column(float64_dtype, missing_value=np.nan)
    insiderPercent = Column(float64_dtype, missing_value=np.nan)
    shortRatio = Column(float64_dtype, missing_value=np.nan)
    year5ChangePercent = Column(float64_dtype, missing_value=np.nan)
    year2ChangePercent = Column(float64_dtype, missing_value=np.nan)
    year1ChangePercent = Column(float64_dtype, missing_value=np.nan)
    ytdChangePercent = Column(float64_dtype, missing_value=np.nan)
    month6ChangePercent = Column(float64_dtype, missing_value=np.nan)
    month3ChangePercent = Column(float64_dtype, missing_value=np.nan)
    month1ChangePercent = Column(float64_dtype, missing_value=np.nan)
    day5ChangePercent = Column(float64_dtype, missing_value=np.nan)

    _loader = IEXKeyStatsLoader()

    @classmethod
    def get_loader(cls):
        return cls._loader


class IEXCompany(DataSet):

    symbol = Column(object_dtype, missing_value='')
    companyName = Column(object_dtype, missing_value='')
    exchange = Column(object_dtype, missing_value='')
    industry = Column(object_dtype, missing_value='')
    website = Column(object_dtype, missing_value='')
    description = Column(object_dtype, missing_value='')
    CEO = Column(object_dtype, missing_value='')
    issueType = Column(object_dtype, missing_value='')
    sector = Column(object_dtype, missing_value='')
    # tags = Column(object_dtype, missing_value='')

    _loader = IEXCompanyLoader()

    @classmethod
    def get_loader(cls):
        return cls._loader



import numpy as np
import pandas as pd

from pipeline_live.data.iex.fundamentals import IEXKeyStats
from pipeline_live.data.iex.pricing import USEquityPricing

from .datamock import mock_iex


def test_IEXKeyStats(refdata, stocks, data_path):

    mock_iex.get_available_symbols(refdata)
    mock_iex.get_key_stats(stocks)

    marketcap = IEXKeyStats.marketcap
    loader = marketcap.dataset.get_loader()
    date = pd.Timestamp('2018-08-20', tz='utc')
    out = loader.load_adjusted_array([marketcap], [date], ['AA'], [])

    assert out[marketcap][0][0] != 0.0


def test_pricing_loader(refdata, stocks, data_path):
    mock_iex.get_available_symbols(refdata)
    mock_iex.get_chart(stocks)

    loader = USEquityPricing.get_loader()
    columns = [USEquityPricing.close]
    dates = [pd.Timestamp('2018-08-22', tz='UTC')]
    symbols = ['AA']
    mask = np.zeros((1, 1), dtype='bool')
    out = loader.load_adjusted_array(columns, dates, symbols, mask)

    assert out[USEquityPricing.close]._data.shape == (1, 1)



import numpy as np
import logbook
import pandas as pd

from zipline.lib.adjusted_array import AdjustedArray
from zipline.pipeline.loaders.base import PipelineLoader
from zipline.utils.calendars import get_calendar
from zipline.errors import NoFurtherDataError

from pipeline_live.data.sources import alpaca


log = logbook.Logger(__name__)


class USEquityPricingLoader(PipelineLoader):
    """
    PipelineLoader for US Equity Pricing data
    """

    def __init__(self):
        cal = get_calendar('NYSE')

        self._all_sessions = cal.all_sessions

    def load_adjusted_array(self, columns, dates, symbols, mask):
        # load_adjusted_array is called with dates on which the user's algo
        # will be shown data, which means we need to return the data that would
        # be known at the start of each date.  We assume that the latest data
        # known on day N is the data from day (N - 1), so we shift all query
        # dates back by a day.
        start_date, end_date = _shift_dates(
            self._all_sessions, dates[0], dates[-1], shift=1,
        )

        sessions = self._all_sessions
        sessions = sessions[(sessions >= start_date) & (sessions <= end_date)]

        timedelta = pd.Timestamp.utcnow() - start_date
        chart_range = timedelta.days + 1
        log.info('chart_range={}'.format(chart_range))
        prices = alpaca.get_stockprices(chart_range)

        dfs = []
        for symbol in symbols:
            if symbol not in prices:
                df = pd.DataFrame(
                    {c.name: c.missing_value for c in columns},
                    index=sessions
                )
            else:
                df = prices[symbol]
                df = df.reindex(sessions, method='ffill')
            dfs.append(df)

        raw_arrays = {}
        for c in columns:
            colname = c.name
            raw_arrays[colname] = np.stack([
                df[colname].values for df in dfs
            ], axis=-1)
        out = {}
        for c in columns:
            c_raw = raw_arrays[c.name]
            out[c] = AdjustedArray(
                c_raw.astype(c.dtype),
                {},
                c.missing_value
            )
        return out


def _shift_dates(dates, start_date, end_date, shift):
    try:
        start = dates.get_loc(start_date)
    except KeyError:
        if start_date < dates[0]:
            raise NoFurtherDataError(
                msg=(
                    "Pipeline Query requested data starting on {query_start}, "
                    "but first known date is {calendar_start}"
                ).format(
                    query_start=str(start_date),
                    calendar_start=str(dates[0]),
                )
            )
        else:
            raise ValueError("Query start %s not in calendar" % start_date)

    # Make sure that shifting doesn't push us out of the calendar.
    if start < shift:
        raise NoFurtherDataError(
            msg=(
                "Pipeline Query requested data from {shift}"
                " days before {query_start}, but first known date is only "
                "{start} days earlier."
            ).format(shift=shift, query_start=start_date, start=start),
        )

    try:
        end = dates.get_loc(end_date)
    except KeyError:
        if end_date > dates[-1]:
            raise NoFurtherDataError(
                msg=(
                    "Pipeline Query requesting data up to {query_end}, "
                    "but last known date is {calendar_end}"
                ).format(
                    query_end=end_date,
                    calendar_end=dates[-1],
                )
            )
        else:
            raise ValueError("Query end %s not in calendar" % end_date)
    return dates[start - shift], dates[end - shift]



from pipeline_live.engine import LivePipelineEngine
from pipeline_live.data.iex.pricing import USEquityPricing
from pipeline_live.data.iex.classifiers import Sector
from pipeline_live.data.iex.factors import AverageDollarVolume
from zipline.pipeline import Pipeline


from .datamock import mock_iex


def test_engine(refdata, stocks, data_path):
    def list_symbols():
        return ['A', 'AA']

    mock_iex.get_available_symbols(refdata)
    mock_iex.get_key_stats(stocks)
    mock_iex.get_chart(stocks)

    eng = LivePipelineEngine(list_symbols)
    ADV = AverageDollarVolume(window_length=20,)
    top5 = ADV.top(5, groupby=Sector())
    pipe = Pipeline({
        'top5': top5,
        'close': USEquityPricing.close.latest,
    }, screen=top5)

    df = eng.run_pipeline(pipe)
    assert sorted(df.columns) == ['close', 'top5']

    pipe = Pipeline({
        'close': USEquityPricing.close.latest,
    })

    df = eng.run_pipeline(pipe)
    assert df.shape == (2, 1)
    assert df.close['AA'] > 40



#!/usr/bin/env python

import ast
import re
from setuptools import setup, find_packages

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('pipeline_live/_version.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

with open('README.md') as readme_file:
    README = readme_file.read()

setup(
    name='pipeline-live',
    version=version,
    description='Zipline Pipeline extension for live trade',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Alpaca',
    author_email='oss@alpaca.markets',
    url='https://github.com/alpacahq/pipeline_live',
    keywords='financial,zipline,pipeline,stock,screening,api,trade',
    packages=find_packages(),
    install_requires=[
        'alpaca-trade-api>=0.29',
        'iexfinance>=0.4.1',
        'zipline==1.3.0',
        'numpy==1.16.1',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'flake8',
    ],
    setup_requires=['pytest-runner', 'flake8'],
)



import numpy as np
from zipline.utils.numpy_utils import (
    object_dtype, datetime64D_dtype, float64_dtype,
)
from zipline.pipeline.data.dataset import Column, DataSet

from .fundamentals_loader import PolygonCompanyLoader


class PolygonCompany(DataSet):

    '''
    "logo": "https://s3.polygon.io/logos/aapl/logo.png",
    "exchange": "Nasdaq Global Select",
    "name": "Apple Inc.",
    "symbol": "AAPL",
    "listdate": "2018-08-15",
    "cik": "0000320193",
    "bloomberg": "EQ0010169500001000",
    "figi": "string",
    "lei": "HWUPKR0MPOU8FGXBT394",
    "sic": 3571,
    "country": "us",
    "industry": "Computer Hardware",
    "sector": "Technology",
    "marketcap": 815604985500,
    "employees": 116000,
    "phone": "(408) 996-1010",
    "ceo": "Tim Cook",
    "url": "http://www.apple.com",
    "description": "Apple Inc. designs, manufactures, and markets mobile communication and media devices, personal computers, and portable digital music players to consumers...\n",
    '''  # noqa

    exchange = Column(object_dtype, missing_value='')
    name = Column(object_dtype, missing_value='')
    symbol = Column(object_dtype, missing_value='')
    listdate = Column(
        datetime64D_dtype,
        missing_value=np.datetime64('1970-01-01'))
    cik = Column(object_dtype, missing_value='')
    bloomberg = Column(object_dtype, missing_value='')
    figi = Column(object_dtype, missing_value='')
    lei = Column(object_dtype, missing_value='')
    sic = Column(float64_dtype, missing_value=np.nan)
    country = Column(object_dtype, missing_value='')
    industry = Column(object_dtype, missing_value='')
    sector = Column(object_dtype, missing_value='')
    marketcap = Column(float64_dtype, missing_value=np.nan)
    employees = Column(float64_dtype, missing_value=np.nan)
    phone = Column(object_dtype, missing_value='')
    ceo = Column(object_dtype, missing_value='')
    tags = Column(object_dtype, missing_value='')

    _loader = PolygonCompanyLoader()

    @classmethod
    def get_loader(cls):
        return cls._loader



from pipeline_live.data.sources import alpaca, iex, polygon

from .datamock import mock_iex, mock_tradeapi


def test_alpaca(alpaca_tradeapi, data_path):
    mock_tradeapi.list_assets(alpaca_tradeapi)
    mock_tradeapi.get_barset(alpaca_tradeapi)

    data = alpaca.get_stockprices(2)
    assert data['AA'].iloc[0].close == 172.18


def test_polygon(tradeapi, data_path):
    mock_tradeapi.list_assets(tradeapi)
    mock_tradeapi.list_financials(tradeapi)

    data = polygon.financials()
    assert len(data['AA']) == 5
    assert data['AA'][-1]['totalCash'] > 10e7

    mock_tradeapi.list_companies(tradeapi)
    data = polygon.company()
    assert len(data) == 2
    assert data['AA']['name'] == 'Alcoa Corp'


def test_iex(refdata, stocks, data_path):
    mock_iex.get_available_symbols(refdata)
    mock_iex.get_key_stats(stocks)

    kstats = iex.key_stats()

    assert len(kstats) == 2
    assert kstats['AA']['latestEPSDate'] == '2017-12-30'

    # cached
    kstats = iex.key_stats()
    assert kstats['AA']['returnOnCapital'] is None

    mock_iex.get_financials(stocks)

    financials = iex.financials()
    assert len(financials) == 1
    assert len(financials['AA']) == 4

    ed = iex._ensure_dict(1, ['AA'])
    assert ed['AA'] == 1

    mock_iex.get_chart(stocks)

    data = iex.get_stockprices()
    assert len(data) == 1
    data['AA'].open[1] == 43.2



from zipline.pipeline.data.dataset import Column, DataSet
from zipline.utils.numpy_utils import float64_dtype

from .pricing_loader import USEquityPricingLoader


# In order to use it as a cache key, we have to make it singleton
_loader = USEquityPricingLoader()


class USEquityPricing(DataSet):
    """
    Dataset representing daily trading prices and volumes.
    """
    open = Column(float64_dtype)
    high = Column(float64_dtype)
    low = Column(float64_dtype)
    close = Column(float64_dtype)
    volume = Column(float64_dtype)

    @staticmethod
    def get_loader():
        return _loader
